// Copyright (C) 2017 Elviss Strazdins
// This file is part of the Ouzel engine.

#include "Cache.h"
#include "Engine.h"
#include "Application.h"
#include "graphics/Renderer.h"
#include "graphics/TextureResource.h"
#include "graphics/ShaderResource.h"
#include "scene/ParticleDefinition.h"
#include "scene/SpriteFrame.h"
#include "files/FileSystem.h"

namespace ouzel
{
    Cache::Cache()
    {
    }

    void Cache::clear()
    {
        releaseTextures();
        releaseShaders();
        releaseParticleDefinitions();
        releaseBlendStates();
        releaseSpriteFrames();
        releaseBMFonts();
    }

    void Cache::preloadTexture(const std::string& filename, bool dynamic, bool mipmaps)
    {
        std::map<std::string, std::shared_ptr<graphics::Texture>>::const_iterator i = textures.find(filename);

        if (i == textures.end())
        {
            std::shared_ptr<graphics::Texture> texture = std::make_shared<graphics::Texture>();
            texture->initFromFile(filename, dynamic, mipmaps);

            textures[filename] = texture;
        }
    }

    const std::shared_ptr<graphics::Texture>& Cache::getTexture(const std::string& filename, bool dynamic, bool mipmaps) const
    {
        std::map<std::string, std::shared_ptr<graphics::Texture>>::const_iterator i = textures.find(filename);

        if (i != textures.end())
        {
            return i->second;
        }
        else
        {
            std::shared_ptr<graphics::Texture> result = std::make_shared<graphics::Texture>();
            result->initFromFile(filename, dynamic, mipmaps);

            i = textures.insert(std::make_pair(filename, result)).first;

            return i->second;
        }
    }

    void Cache::setTexture(const std::string& filename, const std::shared_ptr<graphics::Texture>& texture)
    {
        textures[filename] = texture;
    }

    void Cache::releaseTextures()
    {
        for (auto i = textures.begin(); i != textures.end();)
        {
            // don't delete white pixel texture
            if (i->first == graphics::TEXTURE_WHITE_PIXEL)
            {
                ++i;
            }
            else
            {
                i = textures.erase(i);
            }
        }
    }

    const std::shared_ptr<graphics::Shader>& Cache::getShader(const std::string& shaderName) const
    {
        std::map<std::string, std::shared_ptr<graphics::Shader>>::const_iterator i = shaders.find(shaderName);

        if (i != shaders.end())
        {
            return i->second;
        }
        else
        {
            i = shaders.insert(std::make_pair(shaderName, nullptr)).first;

            return i->second;
        }
    }

    void Cache::setShader(const std::string& shaderName, const std::shared_ptr<graphics::Shader>& shader)
    {
        shaders[shaderName] = shader;
    }

    void Cache::releaseShaders()
    {
        for (auto i = shaders.begin(); i != shaders.end();)
        {
            // don't delete default shaders
            if (i->first == graphics::SHADER_COLOR ||
                i->first == graphics::SHADER_TEXTURE)
            {
                ++i;
            }
            else
            {
                i = shaders.erase(i);
            }
        }
    }

    const std::shared_ptr<graphics::BlendState>& Cache::getBlendState(const std::string& blendStateName) const
    {
        std::map<std::string, std::shared_ptr<graphics::BlendState>>::const_iterator i = blendStates.find(blendStateName);

        if (i != blendStates.end())
        {
            return i->second;
        }
        else
        {
            i = blendStates.insert(std::make_pair(blendStateName, nullptr)).first;

            return i->second;
        }
    }

    void Cache::setBlendState(const std::string& blendStateName, const std::shared_ptr<graphics::BlendState>& blendState)
    {
        blendStates[blendStateName] = blendState;
    }

    void Cache::releaseBlendStates()
    {
        for (auto i = blendStates.begin(); i != blendStates.end();)
        {
            // don't delete default blend states
            if (i->first == graphics::BLEND_NO_BLEND ||
                i->first == graphics::BLEND_ADD ||
                i->first == graphics::BLEND_MULTIPLY ||
                i->first == graphics::BLEND_ALPHA)
            {
                ++i;
            }
            else
            {
                i = blendStates.erase(i);
            }
        }
    }

    void Cache::preloadSpriteFrames(const std::string& filename, bool mipmaps,
                                    uint32_t spritesX, uint32_t spritesY,
                                    const Vector2& pivot)
    {
        std::string extension = sharedApplication->getFileSystem()->getExtensionPart(filename);

        std::vector<scene::SpriteFrame> frames;

        if (extension == "json")
        {
            frames = scene::SpriteFrame::loadSpriteFrames(filename, mipmaps);
        }
        else
        {
            std::shared_ptr<graphics::Texture> texture = sharedEngine->getCache()->getTexture(filename, false, mipmaps);

            if (texture)
            {
                Size2 spriteSize = Size2(texture->getSize().v[0] / spritesX, texture->getSize().v[1] / spritesY);

                for (uint32_t x = 0; x < spritesX; ++x)
                {
                    for (uint32_t y = 0; y < spritesY; ++y)
                    {
                        Rectangle rectangle(spriteSize.v[0] * x,
                                            spriteSize.v[1] * y,
                                            spriteSize.v[0],
                                            spriteSize.v[1]);

                        scene::SpriteFrame frame = scene::SpriteFrame(texture, rectangle, false, spriteSize, Vector2(), pivot);
                        frames.push_back(frame);
                    }
                }
            }
        }

        spriteFrames[filename] = frames;
    }

    const std::vector<scene::SpriteFrame>& Cache::getSpriteFrames(const std::string& filename, bool mipmaps,
                                                                  uint32_t spritesX, uint32_t spritesY,
                                                                  const Vector2& pivot) const
    {
        auto i = spriteFrames.find(filename);

        if (i != spriteFrames.end())
        {
            return i->second;
        }
        else
        {
            std::string extension = sharedApplication->getFileSystem()->getExtensionPart(filename);

            std::vector<scene::SpriteFrame> frames;

            if (extension == "json")
            {
                frames = scene::SpriteFrame::loadSpriteFrames(filename, mipmaps);
            }
            else if (spritesX > 0 && spritesY > 0)
            {
                std::shared_ptr<graphics::Texture> texture = sharedEngine->getCache()->getTexture(filename, false, mipmaps);

                if (texture)
                {
                    Size2 spriteSize = Size2(texture->getSize().v[0] / spritesX, texture->getSize().v[1] / spritesY);

                    for (uint32_t x = 0; x < spritesX; ++x)
                    {
                        for (uint32_t y = 0; y < spritesY; ++y)
                        {
                            Rectangle rectangle(spriteSize.v[0] * x,
                                                spriteSize.v[1] * y,
                                                spriteSize.v[0],
                                                spriteSize.v[1]);

                            scene::SpriteFrame frame = scene::SpriteFrame(texture, rectangle, false, spriteSize, Vector2(), pivot);
                            frames.push_back(frame);
                        }
                    }
                }
            }

            i = spriteFrames.insert(std::make_pair(filename, frames)).first;

            return i->second;
        }
    }

    void Cache::setSpriteFrames(const std::string& filename, const std::vector<scene::SpriteFrame>& frames)
    {
        spriteFrames[filename] = frames;
    }

    void Cache::releaseSpriteFrames()
    {
        spriteFrames.clear();
    }

    void Cache::preloadParticleDefinition(const std::string& filename)
    {
        std::map<std::string, scene::ParticleDefinition>::const_iterator i = particleDefinitions.find(filename);

        if (i == particleDefinitions.end())
        {
            particleDefinitions[filename] = scene::ParticleDefinition::loadParticleDefinition(filename);
        }
    }

    const scene::ParticleDefinition& Cache::getParticleDefinition(const std::string& filename) const
    {
        auto i = particleDefinitions.find(filename);

        if (i != particleDefinitions.end())
        {
            return i->second;
        }
        else
        {
            i = particleDefinitions.insert(std::make_pair(filename, scene::ParticleDefinition::loadParticleDefinition(filename))).first;

            return i->second;
        }
    }

    void Cache::releaseParticleDefinitions()
    {
        particleDefinitions.clear();
    }

    void Cache::preloadBMFont(const std::string& filename)
    {
        std::map<std::string, BMFont>::const_iterator i = bmFonts.find(filename);

        if (i == bmFonts.end())
        {
            bmFonts[filename] = BMFont(filename);
        }
    }

    const BMFont& Cache::getBMFont(const std::string& filename) const
    {
        std::map<std::string, BMFont>::const_iterator i = bmFonts.find(filename);

        if (i != bmFonts.end())
        {
            return i->second;
        }
        else
        {
            i = bmFonts.insert(std::make_pair(filename, BMFont(filename))).first;

            return i->second;
        }
    }

    void Cache::releaseBMFonts()
    {
        bmFonts.clear();
    }
}
